# import packages/modules
# external
import os
import pandas as pd
import shutil
# internal
from GaussParse.utils import CheckFileFormat, checkFile, checkDir, ListFiles


class Manager():
    '''
    Manage gaussian log files
    '''

    def __init__(self, src, file_type, sheet_name):
        self.src = src
        self.file_type = file_type
        self.sheet_name = sheet_name
        self.destination_folder_dir = ''
        self.file_extension = ''
        self.file_name = ''

    def collect_files(self):
        '''
        Collect based on file type
        '''
        try:
            # check file format
            # check file/folder
            if checkFile(self.src):
                # read file
                fileDir, fileName, fileExtension = CheckFileFormat(self.src)
                # check extension
                if fileExtension == '.xlsx':
                    self.file_name = fileName
                    self.file_extension = fileExtension
                    self.destination_folder_dir = os.path.join(
                        fileDir, fileName)
                    # collect
                    self.collect_files_from_excel()
                else:
                    raise Exception('file extension is not valid.')
            else:
                raise Exception('file path is not valid.')
        except Exception as e:
            print(e)

    def collect_files_from_excel(self):
        '''
        Collect gaussian files (*.txt/*.log) saved in an excel file as:

        args:
            sheet_name {str}: excel sheet name

        Example:
            excel file contents:
                No | Name | Src
                ----------------------------
                1  | file1 | ../file1.txt
        '''
        try:
            # load excel file
            df = pd.read_excel(self.src, sheet_name=self.sheet_name)
            # df list
            df_list = df.values.tolist()

            # check file
            self.check_file_exists(df_list)

            # copy all gaussian files to destination folder (the folder excel file exists)
            self.copy_files(df_list, self.destination_folder_dir)

            return True

        except Exception as e:
            print(e)

    def check_file_exists(self, df_list):
        '''
        Check the file exists in that directory

        args:
            df_list {list}: df list

        return:
            file_extension {str}: file extensions
        '''
        try:
            # file extension
            file_extensions = []

            for r in df_list:
                # file no
                file_no = r[0]
                # file name
                file_name = r[1]
                # file src
                file_src = (r[2]).strip()

                # check file exists
                if os.path.isfile(file_src):
                    # check file
                    fileDir, fileName, fileExtension = CheckFileFormat(
                        file_src)
                    # add
                    file_extensions.append(fileExtension)
                    # print(f"exists *** {file_no} {file_name} {file_src} *** ")
                else:
                    raise Exception(
                        f"does not exist *** {file_no} {file_name} {file_src} *** ")

                # res
                return file_extensions
        except Exception as e:
            print(e)

    def copy_files(self, df_list, destination_folder_dir):
        '''
        Copy all files from the source folder to the destination folder 

        args:
            df_list {list}: df list
            destination_folder_dir {str}: destination folder
            file_extension {str}: file extension (default: txt)

        '''
        # check destination folder
        if os.path.exists(destination_folder_dir) is False:
            os.makedirs(destination_folder_dir, exist_ok=True)

        total_files_copied = 0
        # copy files
        for r in df_list:
            # file no
            file_no = r[0]
            # file name
            file_name = r[1]
            # file src
            file_src = (r[2]).strip()

            # extract file info
            fileDir, fileName, fileExtension = CheckFileFormat(
                file_src)

            # file name with extension
            destination_file_name = file_name + fileExtension

            # file dir
            source_folder = os.path.dirname(file_src)
            # file name (in dir)
            source_file_name = os.path.basename(file_src)

            # Check if the source folder exists
            if os.path.exists(source_folder):
                # Copy all files from the source folder to the destination folder
                for filename in os.listdir(source_folder):
                    shutil.copy(os.path.join(source_folder, source_file_name),
                                os.path.join(destination_folder_dir, destination_file_name))
                # update total number of files copied
                total_files_copied += 1
            else:
                print('Source folder does not exist.')

        # total number of files copied
        print(
            f"total number of files copied: {total_files_copied} of {len(df_list)}")



# GaussParse version
__version__ = "1.1.0"


# periodic table dict
pt_dict = {'0': {'AtomicNumber': 1, 'Symbol': 'H'},
           '1': {'AtomicNumber': 2, 'Symbol': 'He'},
           '2': {'AtomicNumber': 3, 'Symbol': 'Li'},
           '3': {'AtomicNumber': 4, 'Symbol': 'Be'},
           '4': {'AtomicNumber': 5, 'Symbol': 'B'},
           '5': {'AtomicNumber': 6, 'Symbol': 'C'},
           '6': {'AtomicNumber': 7, 'Symbol': 'N'},
           '7': {'AtomicNumber': 8, 'Symbol': 'O'},
           '8': {'AtomicNumber': 9, 'Symbol': 'F'},
           '9': {'AtomicNumber': 10, 'Symbol': 'Ne'},
           '10': {'AtomicNumber': 11, 'Symbol': 'Na'},
           '11': {'AtomicNumber': 12, 'Symbol': 'Mg'},
           '12': {'AtomicNumber': 13, 'Symbol': 'Al'},
           '13': {'AtomicNumber': 14, 'Symbol': 'Si'},
           '14': {'AtomicNumber': 15, 'Symbol': 'P'},
           '15': {'AtomicNumber': 16, 'Symbol': 'S'},
           '16': {'AtomicNumber': 17, 'Symbol': 'Cl'},
           '17': {'AtomicNumber': 18, 'Symbol': 'Ar'},
           '18': {'AtomicNumber': 19, 'Symbol': 'K'},
           '19': {'AtomicNumber': 20, 'Symbol': 'Ca'},
           '20': {'AtomicNumber': 21, 'Symbol': 'Sc'},
           '21': {'AtomicNumber': 22, 'Symbol': 'Ti'},
           '22': {'AtomicNumber': 23, 'Symbol': 'V'},
           '23': {'AtomicNumber': 24, 'Symbol': 'Cr'},
           '24': {'AtomicNumber': 25, 'Symbol': 'Mn'},
           '25': {'AtomicNumber': 26, 'Symbol': 'Fe'},
           '26': {'AtomicNumber': 27, 'Symbol': 'Co'},
           '27': {'AtomicNumber': 28, 'Symbol': 'Ni'},
           '28': {'AtomicNumber': 29, 'Symbol': 'Cu'},
           '29': {'AtomicNumber': 30, 'Symbol': 'Zn'},
           '30': {'AtomicNumber': 31, 'Symbol': 'Ga'},
           '31': {'AtomicNumber': 32, 'Symbol': 'Ge'},
           '32': {'AtomicNumber': 33, 'Symbol': 'As'},
           '33': {'AtomicNumber': 34, 'Symbol': 'Se'},
           '34': {'AtomicNumber': 35, 'Symbol': 'Br'},
           '35': {'AtomicNumber': 36, 'Symbol': 'Kr'},
           '36': {'AtomicNumber': 37, 'Symbol': 'Rb'},
           '37': {'AtomicNumber': 38, 'Symbol': 'Sr'},
           '38': {'AtomicNumber': 39, 'Symbol': 'Y'},
           '39': {'AtomicNumber': 40, 'Symbol': 'Zr'},
           '40': {'AtomicNumber': 41, 'Symbol': 'Nb'},
           '41': {'AtomicNumber': 42, 'Symbol': 'Mo'},
           '42': {'AtomicNumber': 43, 'Symbol': 'Tc'},
           '43': {'AtomicNumber': 44, 'Symbol': 'Ru'},
           '44': {'AtomicNumber': 45, 'Symbol': 'Rh'},
           '45': {'AtomicNumber': 46, 'Symbol': 'Pd'},
           '46': {'AtomicNumber': 47, 'Symbol': 'Ag'},
           '47': {'AtomicNumber': 48, 'Symbol': 'Cd'},
           '48': {'AtomicNumber': 49, 'Symbol': 'In'},
           '49': {'AtomicNumber': 50, 'Symbol': 'Sn'},
           '50': {'AtomicNumber': 51, 'Symbol': 'Sb'},
           '51': {'AtomicNumber': 52, 'Symbol': 'Te'},
           '52': {'AtomicNumber': 53, 'Symbol': 'I'},
           '53': {'AtomicNumber': 54, 'Symbol': 'Xe'},
           '54': {'AtomicNumber': 55, 'Symbol': 'Cs'},
           '55': {'AtomicNumber': 56, 'Symbol': 'Ba'},
           '56': {'AtomicNumber': 57, 'Symbol': 'La'},
           '57': {'AtomicNumber': 58, 'Symbol': 'Ce'},
           '58': {'AtomicNumber': 59, 'Symbol': 'Pr'},
           '59': {'AtomicNumber': 60, 'Symbol': 'Nd'},
           '60': {'AtomicNumber': 61, 'Symbol': 'Pm'},
           '61': {'AtomicNumber': 62, 'Symbol': 'Sm'},
           '62': {'AtomicNumber': 63, 'Symbol': 'Eu'},
           '63': {'AtomicNumber': 64, 'Symbol': 'Gd'},
           '64': {'AtomicNumber': 65, 'Symbol': 'Tb'},
           '65': {'AtomicNumber': 66, 'Symbol': 'Dy'},
           '66': {'AtomicNumber': 67, 'Symbol': 'Ho'},
           '67': {'AtomicNumber': 68, 'Symbol': 'Er'},
           '68': {'AtomicNumber': 69, 'Symbol': 'Tm'},
           '69': {'AtomicNumber': 70, 'Symbol': 'Yb'},
           '70': {'AtomicNumber': 71, 'Symbol': 'Lu'},
           '71': {'AtomicNumber': 72, 'Symbol': 'Hf'},
           '72': {'AtomicNumber': 73, 'Symbol': 'Ta'},
           '73': {'AtomicNumber': 74, 'Symbol': 'W'},
           '74': {'AtomicNumber': 75, 'Symbol': 'Re'},
           '75': {'AtomicNumber': 76, 'Symbol': 'Os'},
           '76': {'AtomicNumber': 77, 'Symbol': 'Ir'},
           '77': {'AtomicNumber': 78, 'Symbol': 'Pt'},
           '78': {'AtomicNumber': 79, 'Symbol': 'Au'},
           '79': {'AtomicNumber': 80, 'Symbol': 'Hg'},
           '80': {'AtomicNumber': 81, 'Symbol': 'Tl'},
           '81': {'AtomicNumber': 82, 'Symbol': 'Pb'},
           '82': {'AtomicNumber': 83, 'Symbol': 'Bi'},
           '83': {'AtomicNumber': 84, 'Symbol': 'Po'},
           '84': {'AtomicNumber': 85, 'Symbol': 'At'},
           '85': {'AtomicNumber': 86, 'Symbol': 'Rn'},
           '86': {'AtomicNumber': 87, 'Symbol': 'Fr'},
           '87': {'AtomicNumber': 88, 'Symbol': 'Ra'},
           '88': {'AtomicNumber': 89, 'Symbol': 'Ac'},
           '89': {'AtomicNumber': 90, 'Symbol': 'Th'},
           '90': {'AtomicNumber': 91, 'Symbol': 'Pa'},
           '91': {'AtomicNumber': 92, 'Symbol': 'U'},
           '92': {'AtomicNumber': 93, 'Symbol': 'Np'},
           '93': {'AtomicNumber': 94, 'Symbol': 'Pu'},
           '94': {'AtomicNumber': 95, 'Symbol': 'Am'},
           '95': {'AtomicNumber': 96, 'Symbol': 'Cm'},
           '96': {'AtomicNumber': 97, 'Symbol': 'Bk'},
           '97': {'AtomicNumber': 98, 'Symbol': 'Cf'},
           '98': {'AtomicNumber': 99, 'Symbol': 'Es'},
           '99': {'AtomicNumber': 100, 'Symbol': 'Fm'},
           '100': {'AtomicNumber': 101, 'Symbol': 'Md'},
           '101': {'AtomicNumber': 102, 'Symbol': 'No'},
           '102': {'AtomicNumber': 103, 'Symbol': 'Lr'},
           '103': {'AtomicNumber': 104, 'Symbol': 'Rf'},
           '104': {'AtomicNumber': 105, 'Symbol': 'Db'},
           '105': {'AtomicNumber': 106, 'Symbol': 'Sg'},
           '106': {'AtomicNumber': 107, 'Symbol': 'Bh'},
           '107': {'AtomicNumber': 108, 'Symbol': 'Hs'},
           '108': {'AtomicNumber': 109, 'Symbol': 'Mt'},
           '109': {'AtomicNumber': 110, 'Symbol': 'Ds'},
           '110': {'AtomicNumber': 111, 'Symbol': 'Rg'},
           '111': {'AtomicNumber': 112, 'Symbol': 'Cn'},
           '112': {'AtomicNumber': 113, 'Symbol': 'Nh'},
           '113': {'AtomicNumber': 114, 'Symbol': 'Fl'},
           '114': {'AtomicNumber': 115, 'Symbol': 'Mc'},
           '115': {'AtomicNumber': 116, 'Symbol': 'Lv'},
           '116': {'AtomicNumber': 117, 'Symbol': 'Ts'},
           '117': {'AtomicNumber': 118, 'Symbol': 'Og'}}





class errHandler(Exception):
    def __init__(self, *args: object):
        if args:
            self.message = args[0]
        else:
            self.message = None


class errGeneral(Exception):
    def __init__(self, errCode, errMessage):
        super().__init__(self, errMessage)
        self.errCode = errCode
        self.errMessage = errMessage

    def __str__(self):
        print('errGeneralClass')
        if self.errMessage:
            return f'errGeneralClass, {self.errMessage} {self.errCode}, {self.errType()}'
        else:
            return 'errGeneralClass has not raised'
        
    def errType(self):
        if self.errCode == 1:
            return 'errGeneralClass'
        elif self.errCode == 2:
            return 'errGeneralClass'
        elif self.errCode == 3:
            return 'errGeneralClass'
        






class EnergyUnit:
    '''
    energy unit converter
    '''

    def __init__(self, unit):
        self.unit = unit

    @staticmethod
    def hartree_to_kcal_per_mol(value):
        '''
        convert hartree to kcal/mol
        '''
        return value*627.5

    @staticmethod
    def kcal_per_mol_to_hartree(value):
        '''
        convert kcal/mol to hartree
        '''
        return value/627.5

    @staticmethod
    def hartree_to_kJ_per_mol(value):
        '''
        convert hartree to kJ/mol
        '''
        return value*2625.5

    @staticmethod
    def kJ_per_mol_to_hartree(value):
        '''
        convert kJ/mol to hartree
        '''
        return value/2625.5

    @staticmethod
    def hartree_to_eV(value):
        '''
        convert hartree to eV
        '''
        return value*27.211

    @staticmethod
    def eV_to_hartree(value):
        '''
        convert eV to hartree
        '''
        return value/27.211

    @staticmethod
    def hartree_to_inverse_cm(value):
        '''
        convert hartree to inverse cm
        '''
        return value*219474.6

    @staticmethod
    def inverse_cm_to_hartree(value):
        '''
        convert inverse cm to hartree
        '''
        return value/219474.6



# import packages/modules
import numpy as np
import pandas as pd
import os
import re
# internal
from GaussParse.utils import CheckFileFormat, checkFile, checkDir, ListFiles, generateFileName
from GaussParse.config import pt_dict


class Structure():
    '''
    Analyze input orientation
    '''

    def __init__(self, src):
        self.src = src
        # file path
        self.target_dir = ''

    def toTxt(self, file_name):
        '''
        Transform to txt

        args:
            file_name {str}: file name (if it is empty, generate a name)
        '''
        try:
            # check file/folder
            if checkFile(self.src):
                # read file
                fileDir, fileName, fileExtension = CheckFileFormat(self.src)
                # check extension
                if fileExtension == '.log':
                    file_dir = fileDir
                    file = fileName+fileExtension
                    file_list = [file]

            if checkDir(self.src):
                # read folder
                file_dir = self.src
                file_list = ListFiles(file_dir, fileExtension="log")

            # set
            self.target_dir = str(file_dir)

            # file name
            if len(file_name) == 0:
                file_name = generateFileName("molecular-orientation")

            # analyze each file
            analyzed_files = self.AnalyzeFiles(file_dir, file_list)

            # transform to excel
            res, _ = self.save_data_to_txt(analyzed_files, file_name)

            return res
        except Exception as e:
            print(e)

    def toXYZ(self):
        '''
        Transform to xyz format
        '''
        try:
            # check file/folder
            if checkFile(self.src):
                # read file
                fileDir, fileName, fileExtension = CheckFileFormat(self.src)
                # check extension
                if fileExtension == '.txt':
                    file_dir = fileDir
                    file = fileName+fileExtension
                    file_list = [file]

            if checkDir(self.src):
                # read folder
                file_dir = self.src
                file_list = ListFiles(file_dir)

            # set
            self.target_dir = str(file_dir)

            # analyze each file
            res = self.AnalyzeTxtFiles(file_dir, file_list)

            # return
            return res
        except Exception as e:
            print(e)

    def ReadXYZLogFile(self, filePath, pt_dict):
        '''
        read the content of log file

        args:
            filePath {str}: full name of file with directory and format
            pt_dict {dict}: parodic table dict

        return:
            res {dict}: dictionary data
        '''
        try:
            # file info
            fileDir, fileName, fileExtension = CheckFileFormat(filePath)

            # dict
            item_loc = []
            item_conv = []
            column_names = ['Center Number', 'Atomic Number',
                            'Atomic Symbol', 'X', 'Y', 'Z']

            # index
            k = 1

            # file open
            with open(filePath, "r") as f:
                # find
                fContent = f.read()

                # find
                # ---------------------------------------------------------------------
                # Center     Atomic      Atomic             Coordinates (Angstroms)
                # Number     Number       Type             X           Y           Z
                # ---------------------------------------------------------------------
                # str_sub_1 = 'Number     Number              X              Y              Z'
                str_sub_1 = 'Center     Atomic      Atomic             Coordinates (Angstroms)'
                i_sub_1 = fContent.rfind(str_sub_1)
                str_1 = fContent[i_sub_1:]

                # split str by lines
                str_sub_lines = str_1.splitlines()

                # coordinate
                coordinate_1 = str_sub_lines[1]

                # separator
                sep_1 = str_sub_lines[2]

                # find the second separator
                k = 0
                for item in str_sub_lines:
                    # set regex
                    _line = item

                    # find
                    # force - groups(5)
                    # _res = re.search(r"\s+(\d+)\s+(\d+)\s+([-+]?\d*\.\d+)\s+([-+]?\d*\.\d+)\s+([-+]?\d*\.\d+)", _line)
                    # coordinate - groups(6)
                    _res = re.search(
                        r"\s+(\d+)\s+(\d+)\s+(\w+)\s+([-+]?\d*\.\d+)\s+([-+]?\d*\.\d+)\s+([-+]?\d*\.\d+)", _line)
                    # regex 1
                    # \s+(\d+)\s+(\d+)\s+(\d+)\s+([-+]?\d*\.\d+)\s+([-+]?\d*\.\d+)\s+([-+]?\d*\.\d+)
                    # regex 2
                    # \s+(\d+)\s+(\d+)\s+(\w+)\s+([-+]?\d*\.\d+)\s+([-+]?\d*\.\d+)\s+([-+]?\d*\.\d+)
                    # check
                    if _res:
                        _center_number = _res.group(1)
                        _atomic_number = _res.group(2)
                        _atomic_type = _res.group(3)
                        _X = _res.group(4)
                        _Y = _res.group(5)
                        _Z = _res.group(6)

                        # atomic symbol
                        _atomic_symbol = pt_dict.get(
                            str(int(_atomic_number)-1))['Symbol']
                        # store
                        _ele = {
                            'Center Number': _center_number,
                            'Atomic Number': _atomic_number,
                            'Atomic Symbol': _atomic_symbol,
                            'X': _X,
                            'Y': _Y,
                            'Z': _Z
                        }

                        item_loc.append(_ele)

                    if k > 2 and item == sep_1:
                        sep_2 = item
                        break

                    # update
                    k += 1

            return fileName, item_conv, item_loc, column_names
        except Exception as e:
            raise Exception(e)

    def AnalyzeFiles(self, targetPath, fileList):
        '''
        analyze each file

        args:
            targetPath {str}: target folder
            fileList {list[str]}: list of selected files

        output:
            res {dict}: dictionary of input orientation data
        '''
        try:
            # check
            if len(fileList) > 0:
                # res
                res = []
                for item in fileList:
                    # file path
                    _file_full_path = os.path.join(str(targetPath), str(item))
                    # read file
                    _res = self.ReadXYZLogFile(_file_full_path, pt_dict)
                    # save
                    res.append(_res)

                # return
                return res
            else:
                raise Exception("file list is empty!")
        except Exception as e:
            print(e)

    def save_data_to_txt(self, d, file_name, df_format='string'):
        '''
        save xyz coordination output.log to txt file such as Notepad

        args:
            d {list}: input data
            file name {str}: file name, default: res.txt
        '''

        # text file path
        _file_name = file_name + '.txt'
        # text_file_path = os.getcwd() + '/' + file_name + '.txt'
        text_file_path = os.path.join(self.target_dir, _file_name)

        # open the text file in write mode
        with open(text_file_path, 'w') as f:
            # add each df to the text file
            for item in d:
                # sheet name
                file_name_full = item[0]
                item_conv = item[1]
                item_loc = item[2]
                column_names = item[3]

                # df
                df = pd.DataFrame.from_dict(item_loc)

                # check df format
                if df_format == 'string':
                    # df to string
                    df_string = df.to_string(
                        index=False, col_space=10, justify='left')
                    # Add 6 spaces indent to each line
                    indented_df_string = '\n'.join(
                        ['      ' + line for line in df_string.split('\n')])

                    # sep lin
                    sep_line = '---------------------------------------------------------------------------------'
                    # Split the string into lines
                    lines = indented_df_string.split('\n')

                    # find atomic symbol
                    # atomic_numbers = [item.split(" ")[0] for item in lines]

                    # Find the maximum length of the lines
                    max_length = max(len(line) for line in lines)

                    # write file name as heading
                    # f.write('-' * max_length + '\n')
                    f.write("Input orientation: "+file_name_full+'\n')
                    # f.write('-' * max_length + '\n')

                    # write df string to the text file
                    # header line
                    # f.write(lines[0] + '\n')

                    _header = """---------------------------------------------------------------------------------
Center      Atomic      Atomic                   Coordinates (Angstroms)
Number      Number      Symbol                  X           Y           Z
---------------------------------------------------------------------------------
"""
                    f.write(_header)

                    # content
                    f.write(indented_df_string[max_length+1:] + '\n')

                    # last line
                    f.write(sep_line + '\n\n')
                else:
                    # write file name as heading
                    f.write(file_name_full + '\n')
                    f.write('-' * len(file_name_full) + '\n')

                    # write the header
                    f.write('\t'.join(df.columns) + '\n')

                    # write the rest of the data frame
                    for i in range(df.shape[0]):
                        f.write('\t'.join(map(str, df.iloc[i])) + '\n')
                    f.write('\n')
        # res
        return (True, text_file_path)

    def AnalyzeTxtFiles(self, targetPath, fileList):
        '''
        analyze each file

        args:
            targetPath {str}: target folder
            fileList {list[str]}: list of selected files

        output:
            res {dict}: dictionary of input orientation data
        '''
        try:
            # check
            if len(fileList) > 0:
                # res
                res = []
                for item in fileList:
                    # file path
                    _file_full_path = os.path.join(str(targetPath), str(item))
                    # read file
                    _res = self.SaveToXYZ(_file_full_path)
                    # save
                    res.append(_res)

                # return
                return res
            else:
                raise Exception("file list is empty!")
        except Exception as e:
            print(e)

    def generate_xyz_file(self, atom_list, file_name="molecule"):
        """
        Generates an XYZ file from a list of atom positions with nicely formatted columns.

        args:
        atom_list {list of tuples}: Each tuple contains an element symbol and its x, y, z coordinates.
        file_name {str}: The name of the output XYZ file.

        return:
        file name

        Example input for methane (CH4):
        atom_list = [
            ("C", 0.000000, 0.000000, 0.000000),
            ("H", 0.000000, 0.000000, 1.089000),
            ("H", 1.026719, 0.000000, -0.363000),
            ("H", -0.513360, -0.889165, -0.363000),
            ("H", -0.513360, 0.889165, -0.363000)
        ]
        """
        # number of atoms
        num_atoms = len(atom_list)
        # file
        _file = file_name+".xyz"
        # save
        with open(_file, 'w') as file:
            file.write(f"{num_atoms}\n")
            file.write(f"{file_name}\n")
            for atom in atom_list:
                element, x, y, z = atom
                # file.write(f"{element}\t{x:.6f}\t{y:.6f}\t{z:.6f}\n")
                file.write(
                    f"{element:<2} {x:> 10.6f} {y:> 10.6f} {z:> 10.6f}\n")
        # file name
        return _file

    def SaveToXYZ(self, file_path):
        '''
        read the content of txt file

        args:
            file_path {str}: full name of file with directory and format

        return:
            res {dict}: dictionary data
        '''
        try:
            # file info
            fileDir, fileName, fileExtension = CheckFileFormat(file_path)

            # dict
            item_loc = []
            item_conv = []
            column_names = ['Center Number', 'Atomic Number',
                            'Atomic Symbol', 'X', 'Y', 'Z']

            # index
            k = 1

            # file open
            with open(file_path, "r") as f:
                # find
                fContent = f.read()

                # lines
                lines = fContent.split('\n')

                # title
                title = lines[0].split(":")[1].strip()
                sep_1 = lines[1]

                # split str by lines
                str_sub_lines = lines[4:]

                # find the second separator
                k = 0
                for item in str_sub_lines:
                    # set regex
                    _line = item

                    # find
                    # force - groups(5)
                    # _res = re.search(r"\s+(\d+)\s+(\d+)\s+([-+]?\d*\.\d+)\s+([-+]?\d*\.\d+)\s+([-+]?\d*\.\d+)", _line)
                    # coordinate - groups(6)
                    _res = re.search(
                        r"\s+(\d+)\s+(\d+)\s+(\w+)\s+([-+]?\d*\.\d+)\s+([-+]?\d*\.\d+)\s+([-+]?\d*\.\d+)", _line)
                    # regex 1
                    # \s+(\d+)\s+(\d+)\s+(\d+)\s+([-+]?\d*\.\d+)\s+([-+]?\d*\.\d+)\s+([-+]?\d*\.\d+)
                    # regex 2
                    # \s+(\d+)\s+(\d+)\s+(\w+)\s+([-+]?\d*\.\d+)\s+([-+]?\d*\.\d+)\s+([-+]?\d*\.\d+)
                    # check
                    if _res:
                        _center_number = _res.group(1)
                        _atomic_number = _res.group(2)
                        _atomic_symbol = _res.group(3)
                        _X = _res.group(4)
                        _Y = _res.group(5)
                        _Z = _res.group(6)

                        # store
                        _ele = (_atomic_symbol, float(
                            _X), float(_Y), float(_Z))

                        item_loc.append(_ele)
                    # else:
                    #   print(f"{_res},{k}")

                    if k > 2 and item.strip() == sep_1:
                        sep_2 = item.strip()
                        break

                    # update
                    k += 1

            # save xyz file
            _file_name = self.generate_xyz_file(item_loc, title)

            return _file_name, item_conv, item_loc, column_names
        except Exception as e:
            print(e)

    def FindStructures(self, txt_file, display_res=False):
        '''
        find structures saved in a text file (.txt)

        args:
            txt_file {str}: text file

        return:
            str list
        '''
        pattern = re.compile(
            r'((\w+)?\n)?'
            r'---------------------------------------------------------------------------------\n'
            r' Center      Atomic      Atomic                   Coordinates \(Angstroms\)\n'
            r' Number      Number      Symbol                  X           Y           Z\n'
            r'---------------------------------------------------------------------------------\n'
            r'(.*?)\n---------------------------------------------------------------------------------', re.DOTALL)

        # Find all matches
        matches = pattern.findall(txt_file)

        # res
        res = []

        # Print each match with the name (if present)
        for match in matches:
            full_match, name, coordinates = match
            name = name if name else "No name"
            _record = [name, coordinates]
            res.append(_record)
            # check
            if display_res:
                print(f"Input orientation: {name}")
                print(coordinates)
                print('---')
        # res
        return res



from GaussParse.docs.summary_result import SummaryResult
from GaussParse.docs.structure import Structure
from GaussParse.docs.manager import Manager
from GaussParse.docs.plot_result import PlotResult
from GaussParse.docs.irc_result import IRCResult



# import packages/modules
# external
import os
# internal
from GaussParse.docs import SummaryResult, Structure, Manager, PlotResult, IRCResult
from GaussParse.config import __version__


def main():
    '''
    print app version and description
    '''
    # version
    print(
        f'GaussParse {__version__} is a python package to parse Gaussian output files.')


def app_dir():
    '''
    return app directory
    '''
    return os.path.dirname(os.path.abspath(__file__))


def collect_files_from(file_path, file_type="Excel", sheet_name='Sheet1', ):
    '''
    Collect all gaussian files in a folder, excel file format as:

    Excel file (gaussian log file list.xlsx) located in tests folder

        No | Name | Src
        ----------------------------
        1  | file1 | ../file1.txt

        or

        No | Name | Src
        ----------------------------
        1  | file1 | ../file1.log

    args:
        file_path {str}: file path
        file_type {str}: file type (default: Excel)
        sheet_name {str}: sheet name (default: Sheet1)

    return:
        copy all files listed in the source file to a folder
    '''
    try:
        ManagerClass = Manager(file_path, file_type, sheet_name)
        return ManagerClass.collect_files()
    except Exception as e:
        print(e)


def result_summary_to_excel(src: str):
    '''
    Convert Gaussian results summary text file to excel file, 

    Results -> Summary (Overview, Thermo, Opt) -> save data (txt file)

    args:
        src {str}: file | folder
            file format: *.txt
            folder: all *.txt files in the folder

    return:
        excel file - return value is True
    '''
    try:
        SummaryResultClass = SummaryResult(src)
        return SummaryResultClass.toExcel()
    except Exception as e:
        print(e)


def input_orientation_to_txt(src: str, file_name=""):
    '''
    Save input orientations shown in Gaussian files to a text file,

    The Gaussian log file is parsed to extract data, such as the molecular orientation in the x, y, and z coordinates

    args:
        src {str}: file | folder
            Gaussian file: *.log
            folder: all Gaussian *.log files in the folder

    return:
        txt file {str} - return value is True
    '''
    try:
        StructureClass = Structure(src)
        return StructureClass.toTxt(file_name)
    except Exception as e:
        print(e)


def txt_orientation_to_xyz(src: str):
    '''
    Load an input orientation text file and then transform it to xyz format

    args:
        src {str}: file | folder
            text file: *.txt
            folder: all text *.txt files in the folder

    return:
        xyz file {str} - return value is True
    '''
    try:
        StructureClass = Structure(src)
        return StructureClass.toXYZ()
    except Exception as e:
        print(e)


def plot_energy_profile(file_path, options={}, sheet_name='Sheet1', save_img=True):
    '''
    Plot energy profile from data stored in an excel file,

    Excel file (energy.xlsx) located in tests folder

        Excel file columns including (LABEL, X, Y, GROUP, LEGEND, COLOR, LABEL_POSITION, LABEL_DISPLAY, Y_POSITION, Y_DISPLAY):

        LABEL | X | Y | GROUP | LEGEND | COLOR | LABEL_POSITION | LABEL_DISPLAY | Y_POSITION | Y_DISPLAY
        --------------------------------------------------------------------------------------------------

        LABEL: label text
        X: reaction coordinate
        Y: Gibbs free energy
        GROUP: indicate data series (id: 1, 2, 3)
        LEGEND: legend for each data series (short txt)
        COLOR: data series color (red, green, brown)
        LABEL_POSITION: label position (top, bottom)
        LABEL_DISPLAY: display data
        Y_POSITION: label position (top, bottom)
        Y_DISPLAY: display data

    args:
        file_path {str}: file path (xls format)
        options {dict}: plot options (default)
                plot_type {str}: plot type, linear or gaussian (default: linear)
                img_name {str}: plot name (file name), (default: plot)
                target_dir {str}: plot img to target dir, (default: the same as source file)
                plot_sub {dict}: subplot details (not used)
                row {number}: plot row (not used)
                col {number}: plot column (not used)
                y_label {str}: y label text (default: "Gibbs free energy (kcal/mol)")
                x_label {str}: x label text (default: "Reaction coordinate")
                plot_title {str}: plot title (default: "")
                xlim {list[number]}: x-axis range (default: [10, 10])
                ylim {list[number]}: y-axis range (default: [5, 5])
                figsize {list[number]}: fig size (default: [12, 6])
                label_margin {number}: label margin (default: 4) 
                y_major_locator {number}: set y-axis major locator (default: 5)
        sheet_name {str} : sheet name (default: Sheet1)

    return:
        display energy plot and save it in file directory
    '''
    try:
        PlotResultClass = PlotResult(file_path, sheet_name)
        # check manual options
        if len(options) > 0:
            res = PlotResultClass.init_plot(
                manual_options=options, save_img=save_img)
        else:
            res = PlotResultClass.init_plot(save_img=save_img)
        # return
        return res
    except Exception as e:
        print(e)


def plot_irc_profile(file_path: str, options: dict = {}):
    '''
    Plot IRC profile from gaussian irc log file

    In Gaussian, Results -> IRC/Path... display IRC profile, this figure can be plotted.

    args:
        file_path {str}: file path (log format)
        options {dict}: plot options 
            img_name {str}: plot name (file name), (default: plot)
            target_dir {str}: plot img to target dir, (default: the same as source file)
            y_label {str}: y label text (default: "Gibbs free energy (kcal/mol)")
            x_label {str}: x label text (default: "Reaction coordinate")
            xlim {list[number]}: x-axis range (default: [10, 10])
            ylim {list[number]}: y-axis range (default: [5, 5])
            figsize {list[number]}: fig size (default: [12, 6])
            label_margin {number}: label margin (default: 4) 
            y_major_locator {number}: set y-axis major locator (default: 5)
            plt_style {str}: plot theme (default: bmh),
            line_color {str}: plot line color (default: blue),
            y_unit {str}: y label unit (default: Hartree)

    return:
        save IRC plot
    '''
    try:
        IRCResultClass = IRCResult(file_path)
        # check
        if len(options) > 0:
            res = IRCResultClass.toImage(manual_options=options)
        else:
            res = IRCResultClass.toImage()
        # res
        return res
    except Exception as e:
        print(e)


if __name__ == "__main__":
    main()



# import packages/modules
# external
import os
import datetime
# local
from GaussParse.handlers import errGeneral

def CheckFileFormat(file_path):
    '''
    check file format

    args:
        filePath: file name dir

    return:
        file directory, file name, file format
    '''
    # check file exist
    if os.path.isfile(file_path):
        # file analysis
        fileDir = os.path.dirname(file_path)
        fileNameWithExtension = os.path.basename(file_path)
        fileName, fileExtension = os.path.splitext(fileNameWithExtension)

        # res
        return fileDir, fileName, fileExtension
    else:
        raise Exception('file path is not valid.')
    

def ListFiles(targetPath, fileExtension='txt'):
    '''
    list files in a target file

    args:
        targetPath: target path
        fileExtension: file extension, default is txt, can be changed log

    return:
        a list of files in the target path
    '''
    try:
        # check
        if os.path.exists(targetPath):
            filesFound = []
            for f in os.listdir(targetPath):
                if f.endswith('.'+str(fileExtension)):
                    filesFound.append(f)
            # res
            return filesFound
        else:
            raise Exception("target path is not valid.")

    except Exception as e:
        raise e

def checkFile(file):
    if os.path.isfile(file):
        return True
    else:
        return False

def checkDir(dir):
    if os.path.isdir(dir):
        return True
    else:
        return False

def checkPath(path):
    if os.path.exists(path):
        return True
    else:
        return False


def generateFileName(name):
    '''
    generate a file name
    
    args:
        name: name of the file (without extension)

    return:
        a file name
    '''
    try:
        # check name
        if " " in name:
            raise errGeneral(errCode=1, errMessage="Name cannot contain spaces.")
        
        # get current time
        current_time = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
        # Concatenate the parts to form the filename
        filename = f"{name}_{current_time}"
        return filename
    except errGeneral as e:
        print(e)


# import packages/modules
# external
import os
import pandas as pd
import shutil
import matplotlib.pyplot as plt
from matplotlib import ticker
# internal
from GaussParse.utils import CheckFileFormat, checkFile, checkDir, ListFiles


class PlotResult():
    '''
    Plot results
    '''

    # plot options
    options = {
        "plot_type": "linear",
        "fig_size": [12, 6],
        "img_name": 'plot',
        "target_dir": "",
        "plot_sub": {},
        "y_label": "Gibbs free energy (kcal/mol)",
        "x_label": "Reaction coordinate",
        "plot_title": "",
        "xlim": [10, 10],
        "ylim": [5, 5],
        "label_margin": 4,
        "y_major_locator": 5,
    }

    def __init__(self, src, sheet_name):
        self.src = src
        self.sheet_name = sheet_name

    def init_plot(self, manual_options={}, save_img=True):
        '''
        Plot energy profile
        '''
        try:
            # check file/folder
            if checkFile(self.src):
                # read file
                fileDir, fileName, fileExtension = CheckFileFormat(self.src)
                # update target
                self.options['target_dir'] = fileDir
                # check extension
                if fileExtension == '.xlsx':
                    file_dir = fileDir
                    file = fileName+fileExtension
                else:
                    raise Exception("excel file not found!")

            # update options
            if len(manual_options) > 0:
                self.update_plot_options(manual_options)

            # load excel data
            dfs = self.excelLoader(self.src)
            # plot
            self.mPlot(dfs, save_img=save_img)

            return True
        except Exception as e:
            print(e)

    def update_plot_options(self, data):
        '''
        update plot options

        args:
            data {dict}: plot options
        '''
        for key, value in data.items():
            if key in self.options.keys():
                self.options[key] = value
            else:
                raise Exception("key not found!")

    def setY(self, Y):
        '''
        duplicate Y data

        args:
            Y {list[float]}: data list

        return:
            Y2 {list[float]}: duplicated Y list
        '''
        Y2 = [item for item in Y for _ in range(2)]
        return Y2

    def resetY(self, Y):
        '''
        reset to initial Y data list

        args:
            Y {list[float]}: duplcated data list

        return:
            Y2 {list[float]}: initial Y list
        '''
        Y2 = Y[0::2]
        return Y2

    def setX(self, X, span=10, line=3):
        '''
        change the inital X list as:
            step 1: wider dataset
            step 2: make a platform for each state

        args:
            X {list[int]}: data list
            span {int}: data width
            line {int}: label width

        return:
            X3 {list[int]}: new X data
        '''
        # Step 1: Multiply each element by 10
        X2 = [item * int(span) for item in X]

        # Step 2: For each element in X2, add 1 and then store both original and incremented values
        X3 = []
        for item in X2:
            X3.append(item)
            X3.append(item + int(line))

        return X3

    def resetX(self, X, span=10, line=3):
        '''
        change the inital X list as:
            step 1: wider dataset
            step 2: make a platform for each state

        args:
            X {list[int]}: data list
            span {int}: data width
            line {int}: label width

        return:
            X3 {list[int]}: the initial X data
        '''
        # Step 1: divide each element by 10
        X2 = [item for item in X]
        # step 2: choose even elements
        X3 = X2[0::2]

        return X3

    def excelLoader(self, excel_file):
        '''
        load an excel file containing data as:
        column 1: ID
        column 2: X - coordinate reaction point
        column 3: Y - energy data
        column 4: GROUP - indicate data series
        column 5: LEGEND - legend for each data series
        column 6: COLOR - data series color
        column 7: LABEL_POSITION - label position
        column 8: LABEL_DISPLAY - display data
        column 9: Y_POSITION - label position
        column 10: Y_DISPLAY - display data

        args:
        excel_file {str}: excel file directory

        return:
        data for plotting
        '''
        try:
            # check
            if os.path.exists(excel_file) and os.path.isfile(excel_file):
                # make a dataframe
                df = pd.read_excel(excel_file)
                # get all group values
                groups = list(set(df['GROUP'].tolist()))
                # df list
                dfs = []
                for i in groups:
                    _df = df[df['GROUP'] == i]
                    dfs.append(_df)
                    # rest
                    _df = _df.drop(_df.index)

                data = []
                for item in dfs:
                    # label
                    _LABEL = item['LABEL'].tolist()
                    # X
                    _X = item['X'].tolist()
                    _X2 = self.setX(_X)
                    # Y
                    _Y = item['Y'].tolist()
                    _Y2 = self.setY(_Y)
                    # group
                    _GROUP = list(set(item['GROUP'].tolist()))
                    # legend
                    _iLegend = item.dropna(subset=['LEGEND'])
                    # check
                    if _iLegend.shape[0] == 1:
                        _LEGEND = _iLegend['LEGEND'].tolist()
                    else:
                        print("err")
                    # color
                    _iColor = item.dropna(subset=['COLOR'])
                    # check
                    if _iColor.shape[0] == 1:
                        _COLOR = _iColor['COLOR'].tolist()
                    else:
                        print("err")
                    # label position
                    _iLabelPosition = item.dropna(subset=['LABEL_POSITION'])
                    if _iLabelPosition.shape[0] == 1:
                        _LABEL_POSITION = _iLabelPosition['LABEL_POSITION'].tolist(
                        )
                    else:
                        # default
                        _LABEL_POSITION = ['top']
                    # label display
                    _LABEL_DISPLAY = item['LABEL_DISPLAY'].fillna(
                        "Yes").tolist()
                    # Y position
                    _iYPosition = item.dropna(subset=['Y_POSITION'])
                    if _iYPosition.shape[0] == 1:
                        _Y_POSITION = _iYPosition['Y_POSITION'].tolist()
                    else:
                        # default
                        _Y_POSITION = ['top']
                    # Y display
                    _Y_DISPLAY = item['Y_DISPLAY'].fillna("Yes").tolist()
                    # set data
                    _data = {
                        "LABEL": _LABEL,
                        "X": _X2,
                        "Y": _Y2,
                        "GROUP": _GROUP,
                        "LEGEND": _LEGEND,
                        "COLOR": _COLOR,
                        "LABEL_POSITION": _LABEL_POSITION,
                        "LABEL_DISPLAY": _LABEL_DISPLAY,
                        "Y_POSITION": _Y_POSITION,
                        "Y_DISPLAY": _Y_DISPLAY,
                    }
                    # save
                    data.append(_data)
                    # reset
                    _data = {}

            # res
            return data
        except Exception as e:
            print(e)

    def setPosition(self, position, Xs, Ys, margin=2.5):
        '''
        config text positio to show label

        args:
            position {str}: label position (top/bottom)
            Xs {list[number]}: X values
            Ys {list[number]}: Y values
            margin {number}: move text vertically (default value: 2.5)

        return:
            x_mid, y_mid {list[number]}: new position
        '''
        # set
        x_mid = 0.5 * (Xs[0] + Xs[1])
        y_mid = 0.5 * (Ys[0] + Ys[1])
        margin = 2.5  # Adjust this value as needed for margin
        # check
        if position == 'top':
            y_mid += margin
        elif position == 'bottom':
            y_mid -= margin
        # res
        return x_mid, y_mid

    def mPlot(self, data, save_img):
        '''
        plot data series

        args:
            data: list of dict
            save_img {bool}: set to save plot (default: True)

            other options
                X {list}: X values
                Y {list}: Y values
                LABEL {list}: labels
                LEGEND {list}: legend
                GROUP {list}: group id
                COLOR {list}: data series color
                LABEL_POSITION {list}: set label position on each line (default: top)
                LABEL_DISPLAY {list}: set label display (Yes/No)
                Y_POSITION {list}: set label position on each line (default: top)
                Y_DISPLAY {list}: set label display (Yes/No)
                plot_type {str}: plot type, linear or gaussian
                save_img {bool}: set to save plot (default: False)
                img_name {str}: plot name (file name)
                target_dir {str}: plot img to target dir
                plot_sub {dict}: subplot details
                row {number}: plot row
                col {number}: plot column
                y_label {str}: y label text
                x_label {str}: x label text
                plot_title {str}: plot title
                xlim {list[number]}: x-axis range
                ylim {list[number]}: y-axis range
                figsize {list[number]}: fig size (default: [10,6])
                y_major_locator {number}: set y-axis major locator
        '''
        # set options
        plot_type = self.options.get('plot_type')
        img_name = self.options.get('img_name')
        target_dir = self.options.get('target_dir')
        plot_sub = self.options.get('plot_sub')
        y_label = self.options.get('y_label')
        x_label = self.options.get('x_label')
        plot_title = self.options.get('plot_title')
        xlim = self.options.get('xlim')
        ylim = self.options.get('ylim')
        label_margin = self.options.get('label_margin')
        figsize = self.options.get('fig_size')
        y_major_locator = self.options.get('y_major_locator')

        # create a plot
        fig, ax = plt.subplots(figsize=figsize)

        # check subplot
        if len(plot_sub) > 0:
            print("set subplot")

        # Add a title to the plot
        if len(plot_title) > 0:
            plt.title(str(plot_title), fontweight='bold',
                      color='black', fontsize=16)

        # Set the y-label
        ax.set_ylabel(y_label, fontweight='bold', color='black')
        # set the x-label
        ax.set_xlabel(x_label, fontweight='bold', color='black')

        # Set the label position
        # The x coordinate is typically negative for the y-axis label to be placed outside the plot
        # ax.yaxis.set_label_coords(-0.1, 0.5)
        # ax.xaxis.set_label_coords(0.5, -0.05)

        # Hide the top and right spines
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.spines['bottom'].set_visible(True)

        # Remove the x-axis tick labels
        ax.set_xticklabels([])
        # Hide the x-axis tick symbols
        ax.tick_params(axis='x', which='both', bottom=False,
                       top=False, labelbottom=True)
        # Hide the x-axis
        ax.xaxis.set_visible(True)

        # Set the y-tick labels to display the actual values
        ax.yaxis.set_major_formatter(ticker.StrMethodFormatter('{x:,.2f}'))
        ax.yaxis.set_major_locator(ticker.MultipleLocator(y_major_locator))

        # grid
        # plt.grid(which='both', color='gray', linestyle='--', linewidth=0.5, alpha=0.25)
        # Enable minor ticks
        plt.minorticks_on()
        # Customize major grid lines
        plt.grid(which='major', color='grey',
                 linestyle='-', linewidth=0.5, alpha=0.25)
        # Customize minor grid lines
        plt.grid(which='minor', color='gray',
                 linestyle=':', linewidth=0.4, alpha=0.25)

        # Y min/max
        minY = []
        maxY = []
        # X min/max
        minX = []
        maxX = []

        # plot data series
        for item in data:
            # extract
            X = item['X']
            Y = item['Y']
            labels = item['LABEL']
            legend = str(item['LEGEND'][0])
            color = str(item['COLOR'][0])
            label_position = str(item['LABEL_POSITION'][0])
            label_display = item['LABEL_DISPLAY']
            Y_position = str(item['Y_POSITION'][0])
            Y_display = item['Y_DISPLAY']

            # min
            _minY = min(Y)
            _minX = min(X)
            minY.append(_minY)
            minX.append(_minX)
            # max
            _maxY = max(Y)
            _maxX = max(X)
            maxY.append(_maxY)
            maxX.append(_maxX)

            # plot
            if plot_type == 'linear':
                ax.plot(X, Y, linestyle="dotted",
                        label=legend, color=color, linewidth=1)
            else:
                # gaussian
                _X = self.resetX(X)
                _Y = self.resetY(Y)
                # Create a spline representation
                # spline = UnivariateSpline(_X, _Y, s=0)  # Adjust smoothing factor as needed
                # Generate new data points for a smoother curve
                # _Ynew = spline(_X)
                # plot
                # ax.plot(resetX(X), resetY(Y),"--",label=legend,color=color,linewidth=1)
                # ax.plot(_X,_Ynew,label=legend,color=color,linewidth=1)

            # plot lines
            Xsize = len(X)
            # set labels
            k = 0
            for i in range(0, len(X), 2):
                # set
                # label position
                x_mid, y_mid = self.setPosition(
                    label_position, (X[i], X[i+1]), (Y[i], Y[i+1]), margin=label_margin)
                # Y position
                x_mid_Y, y_mid_Y = self.setPosition(
                    Y_position, (X[i], X[i+1]), (Y[i], Y[i+1]), margin=label_margin)

                # point position
                # ax.scatter(x_mid, y_mid, color='blue', marker='o')
                # ax.scatter(x_mid_Y, y_mid_Y, color='red', marker='o')

                # plot data
                if plot_type == 'linear':
                    ax.plot((X[i], X[i+1]), (Y[i], Y[i+1]),
                            "-", color="black", linewidth=3)
                else:
                    # ax.plot((X[i]-0.5,X[i]+0.5),(Y[i],Y[i]),"-",color="black",linewidth=2)
                    ax.scatter(X[i], Y[i], color='black', marker='o')

                # set
                _Yi_accuracy = "("+str("{:.2f}".format(Y[i]))+")"

                # set labels
                _labels = ''
                # check if the same line
                # check label/Y display
                if label_display[k] == 'Yes' and Y_display[k] == 'Yes':
                    if str(label_position) == str(Y_position):
                        _labels = labels[k]
                        _labels += _Yi_accuracy if len(
                            _labels) == 0 else " "+_Yi_accuracy
                        # set
                        ax.text(x_mid, y_mid, _labels, ha='center',
                                va='center_baseline', fontsize=10)
                    else:
                        ax.text(
                            x_mid, y_mid, labels[k], ha='center', va='center_baseline', fontsize=10)
                        ax.text(x_mid_Y, y_mid_Y, _Yi_accuracy, ha='center',
                                va='center_baseline', fontsize=10)
                else:
                    # check label display
                    if label_display[k] == 'Yes':
                        # set
                        ax.text(
                            x_mid, y_mid, labels[k], ha='center', va='center_baseline', fontsize=10)
                    # check Y display
                    if Y_display[k] == 'Yes':
                        # set
                        ax.text(x_mid_Y, y_mid_Y, _Yi_accuracy, ha='center',
                                va='center_baseline', fontsize=10)
                # set
                k += 1

        # Set the y-axis range
        ax.set_ylim(min(minY)-ylim[0], max(maxY)+ylim[1])
        ax.set_xlim(min(minX)-xlim[0], max(maxX)+xlim[1])

        # Adding a customized legend
        # ax.legend(loc='upper right', fontsize='medium', title='', shadow=True)
        # ax.legend(loc='upper center', bbox_to_anchor=(0.5, -0.05), ncol=2)
        # Adjust x, y for position
        ax.legend(loc='upper left', bbox_to_anchor=(0, 0))

        # theme
        plt.style.use('bmh')

        # plot file name
        img_name = img_name + ".png"
        # check
        if save_img:
            # save a figure
            if len(target_dir) == 0:
                _cwd = os.getcwd()
                img_path = os.path.join(_cwd, img_name)
            else:
                img_path = os.path.join(target_dir, img_name)
            # Apply tight layout for automatic spacing
            plt.tight_layout()
            # Save using the format parameter
            plt.savefig(img_path, facecolor='white',
                        dpi=300, bbox_inches='tight')

        # Display the plot
        plt.show()



from GaussParse.app import result_summary_to_excel, app_dir, input_orientation_to_txt, txt_orientation_to_xyz, collect_files_from, plot_energy_profile, plot_irc_profile
from GaussParse.config import __version__



# import packages/modules
# external
import os
import re
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib import ticker
# internal
from GaussParse.utils import CheckFileFormat, checkFile, checkDir, ListFiles, generateFileName, EnergyUnit


class IRCResult:
    '''
    Parse IRC results
    '''
    # plot options
    options = {
        "fig_size": [12, 6],
        "img_name": 'img',
        "target_dir": "",
        "y_label": "Total Energy",
        "x_label": "Intrinsic Reaction Coordinate",
        "xlim": [10, 10],
        "ylim": [5, 5],
        "label_margin": 4,
        "y_major_locator": 5,
        "plt_style": 'bmh',
        "line_color": "blue",
        "y_unit": "Hartree"
    }

    def __init__(self, src):
        self.src = src

    def toImage(self, manual_options={}):
        '''
        Transform the IRC result to image
        '''
        try:
            # check file/folder
            if checkFile(self.src):
                # read file
                fileDir, fileName, fileExtension = CheckFileFormat(self.src)
                # check extension
                if fileExtension == '.log':
                    file_dir = fileDir
                    file = fileName+fileExtension
                    file_list = [file]

            if checkDir(self.src):
                # read folder
                file_dir = self.src
                file_list = ListFiles(file_dir, fileExtension="log")

            # update options
            if len(manual_options) > 0:
                self.update_plot_options(manual_options)

            # analyze each file
            analyzed_files = self.AnalyzeIRC(file_dir, file_list)

            # plot data and then save
            self.save_data_to_image(analyzed_files, image_file_dir=file_dir)

            return True
        except Exception:
            pass

    def update_plot_options(self, data):
        '''
        update plot options

        args:
            data {dict}: plot options
        '''
        try:
            for key, value in data.items():
                if key in self.options.keys():
                    self.options[str(key)] = str(value)
                else:
                    raise Exception(f"key {key} not found!")
        except Exception as e:
            raise Exception(e)

    def ReadIRC(self, filePath):
        '''
        read the content of IRC log file

        args:
            filePath {str}: full name of file with directory and format

        return:
            res {tuple}
                fileName {str}: file name
                item_loc {dict}: irc data
                column_names {List}: list of column names
        '''
        try:
            # file info
            fileDir, fileName, fileExtension = CheckFileFormat(filePath)

            # dict
            item_loc = []
            column_names = ['X', 'Energy', 'RxCoord']

            # index
            k = 1

            # file open
            with open(filePath, "r") as f:
                # content
                fContent = f.read()
                # find
                str_sub_1 = "Energies reported relative to the TS energy of "
                i_sub_1 = fContent.rfind(str_sub_1)
                str_1 = fContent[i_sub_1:]

                # split str by lines
                str_sub_lines = str_1.splitlines()

                # energy
                E_pattern = r"([-+]?\d*\.\d+|\d+)"
                E_match = re.search(E_pattern, str_sub_lines[0])
                if E_match:
                    E = float(E_match.group())
                    # print(E)
                else:
                    raise Exception(
                        "not found: Energies reported relative to the TS energy")

                # separator
                sep_1 = str_sub_lines[1]
                # table header
                table_header = str_sub_lines[2]

                # find the second separator
                k = 0
                k_set = 4
                for item in str_sub_lines:
                    # set regex
                    _line = item

                    # find
                    _res = re.search(
                        r"(\d+)\s+([-+]?\d*\.\d+|\d+)\s+([-+]?\d*\.\d+|\d+)", _line)
                    # check
                    if _res:
                        _X = _res.group(1)
                        _Energy = _res.group(2)
                        _RxCoord = _res.group(3)
                        # store
                        _ele = {
                            'X': int(_X),
                            'Energy': float(_Energy),
                            'RxCoord': float(_RxCoord),
                            'EnergyTotal': float(_Energy) + E,
                        }

                        item_loc.append(_ele)

                    # check the end of data zone
                    if k > k_set and item == sep_1:
                        sep_2 = item
                        break

                    # update
                    k += 1

            return fileName, item_loc, column_names
        except Exception as e:
            raise Exception(e)

    def AnalyzeIRC(self, targetPath, fileList):
        '''
        Analyze every file in the file list

        args:
            targetPath {str}: target folder
            fileList {List[str]}: list of selected files

        return:
            res: dict
        '''
        try:
            # check
            if len(fileList) > 0:
                # res
                res = []
                for item in fileList:
                    # file path
                    _file_full_path = os.path.join(str(targetPath), str(item))
                    # read file
                    _res = self.ReadIRC(_file_full_path)
                    # save
                    res.append(_res)

                # return
                return res
            else:
                raise Exception("file list is empty!")

        except Exception as e:
            raise Exception(e)

    def save_data_to_image(self, d, image_file_dir=''):
        '''
        Save IRC result output.log to image file format

        args:
            d {List[dict]}: input data
            image_file_dir {str}: file directory used for saving all images

        return:
            status {bool}: True for successful action 
        '''
        # index
        data_len = len(d)

        # unit class
        EnergyUnitClass = EnergyUnit(self.options['y_unit'])

        # add each df to the excel sheet
        for item in d:
            # extract data
            file_name = item[0]
            file_data = item[1]
            file_column_name = item[2]

            # img file name
            img_file_name = file_name.strip()

            #  img file path
            if len(image_file_dir) == 0:
                image_file_dir = os.getcwd()

            # dataframe
            _df = pd.DataFrame(file_data)

            # x values
            _X = _df['RxCoord']
            X = _X.to_numpy()

            # y values
            _Y = _df['EnergyTotal']
            Y = _Y.to_numpy()

            # check unit
            if self.options['y_unit'] == "kcal/mol":
                Y = EnergyUnitClass.hartree_to_kcal_per_mol(Y)
            elif self.options['y_unit'] == "kJ/mol":
                Y = EnergyUnitClass.hartree_to_kJ_per_mol(Y)
            elif self.options['y_unit'] == "eV":
                Y = EnergyUnitClass.hartree_to_eV(Y)
            else:
                # hartree
                Y = Y

                # save image
            self.save_plot(X, Y, file_name, image_file_dir, img_file_name)

            # empty dataframe
            _df = pd.DataFrame()

        # return
        return True

    def save_plot(self, X, Y, title, target_dir, img_file_name, display=False, format="png"):
        '''
        plot irc profile

        args:
            X {list}: x values
            Y {list}: y values
            title {str}: plot title
            target_dir {str}: target directory
            img_file_name {str}: image file name
            display {bool}: display plot (default: False)
            format {str}: image format (default: png, others: svg)
        '''
        # Add a title to the plot
        plt.title(title, fontweight='bold', color='black', fontsize=16)

        # styles
        plt_style = self.options['plt_style']
        plt.style.use(plt_style)

        # axe labels
        plt.xlabel(self.options['x_label'],
                   fontweight='bold', color='black')
        # set y label
        _y_label = self.options['y_label'] + \
            " (" + self.options['y_unit'] + ")"
        plt.ylabel(_y_label, fontweight='bold', color='black')

        # grid
        # plt.grid(which='both', color='gray', linestyle='--', linewidth=0.5)
        # Enable minor ticks
        plt.minorticks_on()
        # Customize major grid lines
        # plt.grid(which='major', color='grey', linestyle='-', linewidth=0.75)
        # Customize minor grid lines
        # plt.grid(which='minor', color='gray', linestyle=':', linewidth=0.5)

        # Customize major grid lines
        plt.grid(which='major', color='grey',
                 linestyle='-', linewidth=0.5, alpha=0.25)
        # Customize minor grid lines
        plt.grid(which='minor', color='gray',
                 linestyle=':', linewidth=0.4, alpha=0.25)

        # line color
        line_color = self.options['line_color']

        # plot data
        plt.plot(X, Y, color=line_color)

        # img file
        img_file = (img_file_name + "." + format).strip()
        # img file path
        img_path = os.path.join(target_dir, img_file)

        # save a figure
        # check
        if format == 'svg':
            # svg
            plt.savefig(img_path, dpi=300, facecolor='white',
                        bbox_inches='tight')
        elif format == 'png':
            # png
            plt.savefig(img_path, facecolor='white',
                        dpi=300, bbox_inches='tight')

        # Customize major tick marks for the current axes
        # plt.gca().tick_params(axis='both', which='major', length=5, width=1, colors='black', direction='in', labelsize=10)
        # Customize minor tick marks for the current axes
        # plt.gca().tick_params(axis='both', which='minor', length=5, width=1, colors='gray', direction='in', labelsize=10)
        # Set the y-axis limits to display the actual values
        plt.gca().yaxis.set_major_formatter(
            ticker.StrMethodFormatter('{x:,.3f}'))
        # Customize the appearance of the spines
        plt.gca().spines['top'].set_color('black')
        plt.gca().spines['right'].set_color('black')
        plt.gca().spines['bottom'].set_color('black')
        plt.gca().spines['left'].set_color('black')

        # Set the minimum and maximum values for the x-axis
        plt.xlim(min(X)-1, max(X)+1)  # Adjust the values as needed

        # check
        if display:
            # show
            plt.show()

        # close
        plt.close()

    def change_unit(self, data):
        '''
        change unit from Hartree to kcal/mol or kJ/mol
        '''
        pass



# import packages/modules
# external
import pandas as pd
import os
import re
# internal
from GaussParse.utils import CheckFileFormat, checkFile, checkDir, ListFiles, generateFileName


class SummaryResult:
    '''
    Parse result class
    '''
    # excel file name
    excel_default_file_name = 'summary_result'

    def __init__(self, src):
        self.src = src

    def toExcel(self):
        '''
        Transform the summary result to excel file

        return:
            True
        '''
        try:
            # check file/folder
            if checkFile(self.src):
                # read file
                fileDir, fileName, fileExtension = CheckFileFormat(self.src)
                # check extension
                if fileExtension == '.txt':
                    file_dir = fileDir
                    file = fileName+fileExtension
                    file_list = [file]

            if checkDir(self.src):
                # read folder
                file_dir = self.src
                file_list = ListFiles(file_dir)

            # analyze each file
            analyzed_files = self.AnalyzeFiles(file_dir, file_list)

            # transform to excel
            self.save_data_to_excel(analyzed_files, excel_file_dir=file_dir)

            return True
        except Exception:
            pass

    def ReadFile(self, filePath):
        '''
        read the content of file and put the info in a matrix

        args:
            filePath: full name of file with directory and format

        return:
            res: matrix of info
        '''
        try:
            # file info
            fileDir, fileName, fileExtension = CheckFileFormat(filePath)

            # dict
            item_rows = {}
            column_names = ['Parameter', 'Value', 'Unit']

            # index
            k = 1

            # file open
            with open(filePath, "r") as f:
                # find main section
                fContent = f.readlines()

                for item in fContent:
                    # to skip empty line
                    if len(item) > 1:
                        # set id
                        _id = "data "+str(k)
                        # find "=" otherwise is description line
                        if item.find("=") > -1:
                            # split: data 1 = data 2
                            data1, data2 = item.strip().split("=", 1)

                            # key
                            _key = str(data1).replace("\n", "").strip()

                            # check "space" between number and string
                            _str_search = data2.strip()
                            if _str_search.find(" ") > -1:
                                # check begins with [-][+][number]
                                if _str_search.startswith("-") or _str_search.startswith("+") or _str_search[0].isdigit():
                                    # sort data
                                    _ext = re.search(
                                        r"([-+]?\d*\.?\d+([eE][-+]?\d+)?)\s*(.*)", _str_search, re.M)
                                    # check
                                    if _ext:
                                        numeric_value = float(_ext.group(1))
                                        unit = str(_ext.group(3))
                                else:
                                    numeric_value = _str_search
                                    unit = ''
                            else:
                                # only one part value (without unit)
                                # check numeric
                                _data2 = str(data2).replace(
                                    "\n", "").replace(".", "").strip()
                                if _data2.isnumeric():
                                    # number
                                    numeric_value = float(data2)
                                else:
                                    # string
                                    numeric_value = data2
                                unit = ""

                        else:
                            # description line
                            _key = str(item).replace("\n", "")
                            numeric_value = ''
                            unit = ''

                        # store
                        item_rows[_id] = {"Parameter": str(
                            _key), "Value": numeric_value, "Unit": str(unit)}

                        # set
                        k += 1

            return fileName, item_rows, column_names
        except Exception as e:
            print(e)

    def AnalyzeFiles(self, targetPath, fileList):
        '''
        analyze each file

        args:
            targetPath: target folder
            fileList: list of selected files

        output:
            res: dict
        '''
        try:
            # check
            if len(fileList) > 0:
                # res
                res = []
                for item in fileList:
                    # file path
                    _file_full_path = os.path.join(str(targetPath), str(item))
                    # read file
                    _res = self.ReadFile(_file_full_path)
                    # save
                    res.append(_res)

                # return
                return res
            else:
                raise Exception("file list is empty!")

        except Exception as e:
            print(e)

    def save_data_to_excel(self, d, excel_file_dir='', excel_file_name='', excel_engine='xlsxwriter'):
        '''
        save gaussian output.log to excel

        args:
            d: input data
                file name
        '''
        # file name
        if excel_file_name == '':
            excel_file_name = generateFileName(self.excel_default_file_name)

        # update name
        excel_file_name = excel_file_name+'.xlsx'
        #  excel file path
        if len(excel_file_dir) > 0:
            # chosen path
            excel_file_path = os.path.join(
                excel_file_dir, excel_file_name)
        else:
            # current directory
            excel_file_path = os.getcwd()+'/'+excel_file_name

        # check file exist
        # if not os.path.isfile(excel_file_path):
        #     if not os.path.exists(excel_file_path):
        #         # create excel file
        #         writer = pd.ExcelWriter(excel_file_path,engine='xlsxwriter')
        #         # writer.close()

        writer = pd.ExcelWriter(excel_file_path, engine='xlsxwriter')

        # add each df to the excel sheet
        for item in d:
            # sheet name
            sheet_name = item[0]
            sheet_data = list(item[1].values())
            sheet_column_name = item[2]

            # df
            df = pd.DataFrame.from_dict(sheet_data)
            # size
            row, col = df.shape

            # check excel engine
            if excel_engine == 'xlsxwriter':
                # store
                df.to_excel(writer, sheet_name=str(sheet_name),
                            index=False, header=True, columns=sheet_column_name)

                # workbook and worksheet objects
                workbook = writer.book
                worksheet = writer.sheets[str(sheet_name)]

                # set column
                worksheet.set_column(0, 0, 110)
                worksheet.set_column(1, 2, 30)

                # set boarder
                border_format = workbook.add_format({'border': 1})
                worksheet.conditional_format(
                    0, 0, row, col, {'type': 'no_blanks', 'format': border_format})
                worksheet.conditional_format(
                    0, 0, row, col, {'type': 'blanks', 'format': border_format})

            else:
                # store
                with pd.ExcelWriter(excel_file_path, engine='openpyxl', mode='a', if_sheet_exists="new") as writer2:
                    try:
                        df.to_excel(writer2, sheet_name=str(
                            sheet_name), index=False, header=True, columns=sheet_column_name)
                    except Exception as e:
                        df.to_excel(writer2, sheet_name=str(
                            sheet_name), index=False)
                        print(e)

        # close
        writer.close()
